import { commandQueue } from "./util/commandQueue";
import { entityIsArchetype } from "./util/entityIsArchetype";
import { idGenerator } from "./util/idGenerator";
import { ListenerRegistry } from "./util/ListenerRegistry";
import { memoizedMap } from "./util/memoizedMap";
/**
 * Create an ECS instance.
 */
export function createWorld() {
    /**
     * An array holding all entities known to this world.
     */
    const entities = new Array();
    let nextId = idGenerator(1);
    /**
     * A register of archetype-specific indices.
     */
    const archetypes = new Map();
    const listeners = {
        archetypeChanged: new Map()
    };
    /**
     * A useful queue of commands to run.
     */
    const queue = commandQueue();
    /**
     * A memoization cache for archetypes to help with maintaining
     * referential integrity.
     */
    const memoizedArchetypes = memoizedMap();
    function createArchetype(...components) {
        const normalized = components.sort();
        /* Note: we're only memoizing to make sure that we're always using the same object
           for any combinations of component names that are equal. */
        const archetype = memoizedArchetypes.fetch(normalized, () => normalized);
        /* Create an index if we need to. */
        if (!archetypes.has(archetype)) {
            // console.debug("[hmecs] Creating archetype index for:", archetype)
            /* create an index, and index existing entities */
            archetypes.set(archetype, entities.filter((entity) => entityIsArchetype(entity, archetype)));
            /* Create listeners */
            listeners.archetypeChanged.set(archetype, new ListenerRegistry());
        }
        return archetype;
    }
    function get(archetype) {
        return archetypes.get(archetype);
    }
    function getOne(archetype) {
        return archetypes.get(archetype)[0];
    }
    function getWith(...components) {
        return get(createArchetype(...components));
    }
    /**
     * A bag of functions that will immediately mutate the ECS world.
     * In most situations, you will probably want to call their
     * non-immediate counterparts.
     */
    const immediately = {
        addEntity: (entity) => {
            /* If there already is an ID, raise an error. */
            if ("id" in entity)
                throw "Attempted to add an entity that aleady had an 'id' component.";
            /* Assign an ID */
            entity.id = nextId();
            /* Store the entity... */
            entities.push(entity);
            /* ...and add it to relevant indices. */
            indexEntityWithNewComponents(entity, Object.keys(entity));
            return entity;
        },
        removeEntity: (entity) => {
            /* Remove entity from all indices */
            removeEntityFromAllIndices(entity);
            /* Remove its id component */
            delete entity.id;
            /* Remove it from our global list of entities */
            const pos = entities.indexOf(entity, 0);
            entities.splice(pos, 1);
        },
        addComponent: (entity, name, data) => {
            entity[name] = data;
            indexEntityWithNewComponents(entity, [name]);
        },
        removeComponent: (entity, ...components) => {
            components.forEach((name) => delete entity[name]);
            indexEntityWithRemovedComponents(entity, components);
        }
    };
    function addEntity(entity) {
        queue.add(() => immediately.addEntity(entity));
        return entity;
    }
    function removeEntity(entity) {
        queue.add(() => immediately.removeEntity(entity));
    }
    function addComponent(entity, name, data) {
        queue.add(() => immediately.addComponent(entity, name, data));
    }
    function removeComponent(entity, ...names) {
        queue.add(() => immediately.removeComponent(entity, ...names));
    }
    function flush() {
        queue.flush();
    }
    function indexEntityWithNewComponents(entity, addedComponents) {
        /* When one or more components are added to an entity, it can only be _added_ to indices;
           and only those indices that are interested in any of the added components. Let's go! */
        for (const [archetype, index] of archetypes.entries()) {
            /* If this archetype is interested in any of the new components, logic dictates
               that this entity can't possibly already be listed, and only then may we be
               interested in adding it. */
            if (archetype.some((indexedComponent) => addedComponents.includes(indexedComponent))) {
                /* Now we know the index is potentially interested in this entity, so let's check! */
                if (entityIsArchetype(entity, archetype)) {
                    index.push(entity);
                    listeners.archetypeChanged.get(archetype).invoke();
                }
            }
        }
    }
    function indexEntityWithRemovedComponents(entity, removedComponents) {
        /* When a component is removed from an entity, logic dictates that the only archetype
           indices that are potentially affected are those interested in any of the removed
           component names, so we only need to check those. */
        for (const [archetype, index] of archetypes.entries()) {
            if (archetype.some((indexedComponent) => removedComponents.includes(indexedComponent))) {
                /* By this point the entity _must_ already be in this index, so let's check if
                   it still matches the archetype, and remove it if it doesn't. */
                if (!entityIsArchetype(entity, archetype)) {
                    index.splice(index.indexOf(entity), 1);
                    listeners.archetypeChanged.get(archetype).invoke();
                }
            }
        }
    }
    function removeEntityFromAllIndices(entity) {
        for (const [archetype, index] of archetypes.entries()) {
            const pos = index.indexOf(entity, 0);
            if (pos >= 0) {
                index.splice(pos, 1);
                listeners.archetypeChanged.get(archetype).invoke();
            }
        }
    }
    function clear() {
        /* Remove all entities */
        entities.length = 0;
        /* Remove all archetype indices */
        archetypes.clear();
        /* Remove all listeners */
        listeners.archetypeChanged.clear();
        /* Clear queue */
        queue.clear();
    }
    return {
        entities,
        listeners,
        createArchetype,
        get,
        getOne,
        getWith,
        addEntity,
        removeEntity,
        addComponent,
        removeComponent,
        immediately,
        flush,
        clear
    };
}
